:- module(mis_grammar,[parse_mis/3,mis_methodName/2],[dcg]).

:- use_module(library(filenames)).
:- use_module(library(atom2term)).
:- use_module(library(lists)).

:- use_module(library('jvm_in_ciao/interpreter/jvml')).

% Parameters = formal parameters
% Arguments = effective parameters

mis_methodName(Mis,MethodName) :-
	atom_codes(Mis,SMis),
	mis_methodName_(MethodName,SMis,"").
mis_methodName_(MethodName) -->
	parse_result(_),
	parse_methodName(methodName(_,shortMethodName(MethodName))),
	parse_class(_).

parse_mis(mis(MethodSignature,Arguments,Heap),D,Program) :-
	Heap = heap(dynamicHeap(DynHeap),staticHeap([])),
	atom_codes(D,SD),
	parse_mis_(MethodSignature,Arguments,DynHeap,Program,SD,"").

parse_mis_(methodSignature(MethodName,Parameters,Result),Arguments,DynHeap,P) --> 
	parse_result(Result), 
	%" ", 
	parse_methodName(MethodName),
	"(", 
	parse_params([],Parameters,[],ArgumentsPre,[],DynHeapPre,P),
	")",
	{init_this_object(P,methodSignature(MethodName,Parameters,Result),
	 ArgumentsPre,Arguments,DynHeapPre,DynHeap)}.

init_this_object(P,MethodSignature,ArgsPre,Args,HeapPre,Heap) :-
	MethodSignature = methodSignature(methodName(ClassName,_),_,_),
	program_class(P,ClassName,Class),
	class_method(Class,MethodSignature,Method),
	Method = method(_,_,_,static(IsStatic),_),
	init_this_object_(P,ClassName,IsStatic,ArgsPre,Args,HeapPre,Heap).

init_this_object_(_,_,true,Args,Args,H,H).
init_this_object_(Program,ClassName,false,ArgsPre,[Arg|ArgsPre],HeapPre,Heap) :-
	update_heap(Program,ClassName,HeapPre,Heap,Arg).

parse_methodName(methodName(ClassName,shortMethodName(SMN))) --> 
	parse_className(ClassName),
	".",
	parse_string(SMN).

parse_className(className(PackageName,shortClassName(SCN))) -->
	parse_packageName(PackageName),
	".",
	parse_string(SCN).
parse_className(className(packageName(''),shortClassName(SCN))) -->
	parse_string(SCN).

parse_packageName(packageName(PackageName)) -->
	parse_string(PackageName).

parse_result(none) --> "void ",!.
parse_result(F) --> parse_field_type(F).

parse_params(TsAcu,Ts,VsAcu,Vs,HsAcu,Hs,P) --> 
	parse_field_type(T), 
	%" ",
	parse_value(V,T,HsAcu,NewHsAcu,P),
	",",
	!,
	parse_params([T|TsAcu],Ts,[V|VsAcu],Vs,NewHsAcu,Hs,P).
parse_params(TsAcu,Ts,VsAcu,Vs,HsAcu,NewHsAcu,P) --> 
	parse_field_type(T), 
	%" ",
	parse_value(V,T,HsAcu,NewHsAcu,P),
	{reverse([T|TsAcu],Ts),reverse([V|VsAcu],Vs)}.
parse_params([],[],[],[],[],[],_) --> "".

parse_value(num(int(V)),primitiveType(int),H,H,_) --> 
	parse_string(S),{atom2term(S,V),integer(V)}.
parse_value(num(int(_)),primitiveType(int),H,H,_) --> "_".
parse_value(ref(loc(N)),refType(arrayType(primitiveType(int))),H,NewH,_) --> 
	"_",{append(H,[array(locationArray(5,primitiveType(int)),_)],NewH),
	     length(NewH,N)}.
parse_value(Ref,refType(classType(ClassName)),H,NewH,Program) --> 
	"_",{update_heap(Program,ClassName,H,NewH,Ref)}.
parse_value(null,refType(_),H,H,_) --> 
	"null".

parse_string(S) --> 
	parse_string_(S1),
	{atom_codes(S,S1)}.
parse_string_([S|Ss]) --> 
	[S],
	{member(S,"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_")},
	!,
	parse_string_(Ss).
parse_string_([]) --> "".

parse_field_type(F) --> parse_base_type(F)," ",!.
parse_field_type(F) --> parse_array_type(F)," ",!.
parse_field_type(F) --> parse_object_type(F)," ".

parse_base_type(primitiveType(byte)) --> "byte".
parse_base_type(primitiveType(char)) --> "char".
parse_base_type(primitiveType(double)) --> "double".
parse_base_type(primitiveType(float)) --> "float".
parse_base_type(primitiveType(int)) --> "int".
parse_base_type(primitiveType(long)) --> "long".
parse_base_type(primitiveType(short)) --> "short".
parse_base_type(primitiveType(boolean)) --> "boolean".

parse_object_type(refType(classType(className(packageName(Package),
 	                                      shortClassName(ClassOnly))))) --> 
 	parse_class(Class),
 	{atom_codes(ClassAtom,Class),
 	no_path_file_name(ClassAtom,ClassOnly),
 	atom_concat(Package,ClassOnly,ClassAtom)}.

parse_array_type(refType(arrayType(F))) --> parse_base_type(F),"[]".

parse_class([C]) --> [C].
parse_class([C|R]) --> [C],parse_class(R).

%I am not handleing yet cases in which an array is a member of an object.
update_heap(Program,ClassName,H,NewH,ref(loc(NPlus1))) :-
	length(H,N),NPlus1 is N+1,
	program_class(Program,ClassName,Class),
	class_fields(Class,Fields),
	append(H,[object(locationObject(ClassName),ObjectFields)],H2),
	update_with_fields(Program,Fields,H2,NewH,[],ObjectFields).

update_with_fields(_,[],H,H,OFR,OF) :- reverse(OFR,OF).
update_with_fields(P,[Field|Fields],H,NewH,OFs,NewOFs) :-
	update_with_field(P,Field,H,HPrime,OF),
	update_with_fields(P,Fields,HPrime,NewH,[OF|OFs],NewOFs).

update_with_field(_,Field,H,H,objectField(Signature,num(Num))) :-
	field_type(Field,primitiveType(Type)),
	field_signature(Field,Signature),
	Num =..[Type,_].
update_with_field(P,Field,H,NewH,objectField(Signature,Ref)) :-
	field_type(Field,refType(classType(ClassName))),
	field_signature(Field,Signature),
	update_heap(P,ClassName,H,NewH,Ref).

field_type(field(fieldSignature(_,Type),_Final,_Static,_Visibility,_Init),Type).
	
%[object(locationObject(className(packageName(''),shortClassName('Rational'))),[objectField(fieldSignature(fieldName(className(packageName(''),shortClassName('Rational')),shortFieldName(num)),primitiveType(int)),num(int(3))),objectField(fieldSignature(fieldName(className(packageName(''),shortClassName('Rational')),shortFieldName(den)),primitiveType(int)),num(int(5)))])]
