:- module(program_loader,[load_program/2,load_program/3,load_program2/3]).

:- use_module(classfile_reader).
:- use_module(library('jvm_in_ciao/builtins/builtins')).
:- use_module(mis_grammar).

:- use_module(library(filenames)).
:- use_module(library(terms),[atom_concat/2]).
:- use_module(library(format),[format/2]).
:- use_module(library(prolog_sys),[statistics/2]).

load_program(Program,MIS) :-
	load_program(Program,MIS,[top,trace]).

load_program(Program,MIS,Options) :-
	statistics(runtime,[Time1,_]),
	mis_methodName(MIS,MethodName),
	prepare_stream(Program,MethodName,FOut,ModuleName),
	write_header(FOut,ModuleName),
	load_classes(Program,FOut,[],ClassTerms1),
	statistics(runtime,[Time2,_]),
	builtins_list(BL),
	load_builtins(FOut,BL,ClassTerms1,ClassTerms2),
	load_source_program(FOut,ClassTerms2,ProgramTerm),
	format(FOut,"\n:- include(library(~q)).\n\n",
	       ['jvm_in_ciao/interpreter/pe_i.pl']),
	load_current_mis(FOut,MIS,ProgramTerm),
	load_current_options(FOut,Options),
	close(FOut),
	statistics(runtime,[Time3,_]),
	print_times(Time1,Time2,Time3).

load_program2(Program,MIS,Options) :-
	statistics(runtime,[Time1,_]),
	mis_methodName(MIS,MethodName),
	prepare_stream(Program,MethodName,FOut,ModuleName),
	%write_header(FOut,ModuleName),
	atom_concat(ModuleName,'_program',ModuleProgName),
	atom_concat(ModuleProgName,'.pl',ProgramFileName),
	format(FOut,":- module(~q,[main/2]).\n\n",[ModuleName]),%
	format(FOut,":- use_module(~q).\n\n",[ModuleProgName]),%
	open(ProgramFileName,write,FOutProg),%
	format(FOutProg,":- module(~q,[bytecode/5,class/2,source_program/1]).\n\n",[ModuleProgName]),%
	format(FOutProg,":- discontiguous bytecode/5, class/2.\n\n",[]),
	load_classes(Program,FOutProg,[],ClassTerms1),
	statistics(runtime,[Time2,_]),
	builtins_list(BL),
	load_builtins(FOutProg,BL,ClassTerms1,ClassTerms2),
	load_source_program(FOutProg,ClassTerms2,ProgramTerm),
	format(FOut,"\n:- include(library(~q)).\n\n",
	       ['jvm_in_ciao/interpreter/pe_i.pl']),
	load_current_mis(FOut,MIS,ProgramTerm),
	load_current_options(FOut,Options),
	close(FOut),
	close(FOutProg),
	statistics(runtime,[Time3,_]),
	print_times(Time1,Time2,Time3).

prepare_stream(Program,MethodName,FOut,ModuleName) :-
	Program = [ClassFile|_],
	no_path_file_name(ClassFile,Filename),
	basename(Filename,Basename),
	atom_concat([Basename,'_',MethodName,'_pe'],ModuleName),
	atom_concat(ModuleName,'.pl',OutputName),
	open(OutputName,write,FOut).

write_header(FOut,ModuleName) :-
	format(FOut,":- module(~q,[main/2],[]).\n\n",[ModuleName]),
	format(FOut,":- discontiguous bytecode/5, class/2.\n\n",[]).

print_times(Time1,Time2,Time3) :-
	ClassesTime is Time2 - Time1,
	BuiltinsTime is Time3 - Time2,
	TotalTime is Time3 - Time1,
	format("Classes loading in ~2f \nBuiltins loading in ~2f\nTotal loading time: ~2f\n",[ClassesTime,BuiltinsTime,TotalTime]).

load_classes([ClassFile1|R],FOut,ClassTermsAcu,ClassTerms) :-
	open(ClassFile1,read,FIn),
	build_options_list(ClassFile1,Opts),
	set_options(Opts),
	read_file(FIn,FOut,ClassTerm1),
	nl(FOut),
	close(FIn),
	load_classes(R,FOut,[ClassTerm1|ClassTermsAcu],ClassTerms).
load_classes([],_,L,L).
	
build_options_list(ClassFile,[f(on),r(on),l(on),o(PathOut),c(on)]) :-
	no_path_file_name(ClassFile,Filename),
	basename(Filename,Basename),
	atom_concat([Basename,'_class.pl'],PathOut).

load_builtins(_,[],L,L).
load_builtins(FOut,[B1|RBs],ClassTermsAcu,ClassTerms) :-
	load_builtin(FOut,B1,ClassTerm),
	load_builtins(FOut,RBs,[ClassTerm|ClassTermsAcu],ClassTerms).

load_builtin(FOut,Builtin,ClassTerm) :-
	load_bytecodes(FOut,Builtin),
	builtins:class(Builtin,ClassTerm),
	portray_clause(FOut,class(Builtin,ClassTerm)),nl(FOut).

load_bytecodes(FOut,Builtin) :-
	builtins:bytecode(Builtin,A,B,C,D),
	portray_clause(FOut,bytecode(Builtin,A,B,C,D)),
	fail.
load_bytecodes(_,_).
	
load_source_program(FOut,ClassTermsR,Program) :-
	reverse(ClassTermsR,ClassTerms),
	Program = program(ClassTerms,[]),
	portray_clause(FOut,source_program(Program)).

load_current_mis(FOut,MIS_J,Program) :-
	parse_mis(MIS_P,MIS_J,Program),
	portray_clause(FOut,current_mis(MIS_P)).

load_current_options(FOut,Options) :-
	portray_clause(FOut,options(Options)).
